/******************************************************************\
* zeniko's Setup - A tiny setup routine for launching .inf files.  *
* Copyleft (C) 2001 - 2003  Simon Bnzli (zeniko@gmx.ch)           *
*                                                                  *
* Distributed under the GNU General Public License (see zetup.htm) *
\******************************************************************/

#include <windows.h>
#include <shellapi.h>
#include <dlgs.h>
#include "zetup.h"

HINSTANCE hInst;
BOOL bQuiet = FALSE, bErrors = FALSE, bUninstall = FALSE;
LPTSTR lpAppTitle = APPTITLE;
TCHAR szResource[MAX_STRING];
UINT wm_LBSelChString;

VOID SetupMain(LPTSTR lpUninstFile);
VOID InstallINF(LPTSTR lpFileName, BOOL bUninst);
BOOL BrowseForFolder(LPTSTR lpInitFolder, LPTSTR lpAppSubFolder);
VOID SelfDestruct(VOID);


INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, INT nCmdShow)
{
	LPTSTR lpMoveOn;
	TCHAR cRemoved;
	
	hInst = hInstance;
	
CheckNextParameter:
/* Check the command line for the next parameter */
	for (; isspace(*lpCmdLine); lpCmdLine++);
	for (lpMoveOn = lpCmdLine; (*lpMoveOn != 0) && !isspace(*lpMoveOn); lpMoveOn++);
	cRemoved = *lpMoveOn;
	*lpMoveOn = 0;
	if (lstrcmpi(lpCmdLine, TEXT("/q")) == 0)
	{ /* /q makes zetup run in Quiet mode (without user interaction) */
		bQuiet = TRUE;
		if (cRemoved != 0)
		{ /* look for the next parameter */
			lpCmdLine = lpMoveOn + 1;
			goto CheckNextParameter;
		}
	}
	else if (lstrcmp(lpCmdLine, TEXT("/u")) == 0)
	{ /* uninstall the first filename in the command line */
		bUninstall = TRUE;
		for (lpCmdLine = lpMoveOn + 1; isspace(*lpCmdLine); lpCmdLine++);
		if (lpCmdLine[0] == '"')
		{ /* search closing quotation mark */
			for (lpMoveOn = lpCmdLine; (*lpMoveOn != '"') && (*lpMoveOn != 0); lpMoveOn++);
		}
		else
		{ /* search end of filename */
			for (lpMoveOn = lpCmdLine; (!isspace(*lpMoveOn)) && (*lpMoveOn != 0); lpMoveOn++);
		}
		*lpMoveOn = 0;
	}
	else if (lstrcmp(lpCmdLine, TEXT("/?")) == 0)
	{ /* display help screen and quit */
		if (!bQuiet)
		{
			LOADRESSTRING(1);
			MessageBox(NULL, szResource, lpAppTitle, MB_OK | MB_ICONINFORMATION);
		}
		bErrors = bQuiet; /* can't display help screen in Quiet mode */
		goto ErrorQuitPoint;
	}
	
	wm_LBSelChString = RegisterWindowMessage(LBSELCHSTRING);
	
	SetupMain(lpCmdLine);
	
ErrorQuitPoint:
/* check the errorlevel when running in Quiet mode */
	return (bErrors && bQuiet)?1:0;
}


VOID SetupMain(LPTSTR lpUninstFile)
{
	TCHAR szFilePath[MAX_PATH], szProfileString[MAX_STRING];
	LPTSTR lpNameOnly, lpQuestion;
	WIN32_FIND_DATA searchData;
	DWORD dwSize = MAX_PATH;
	HANDLE hSearch;
	INT iResult;
	HKEY hKey;
	
	if (!bUninstall)
	{ /* Search for all .inf files in zetup's folder */
		GetModuleFileName(hInst, szFilePath, MAX_PATH);
		GetFullPathName(szFilePath, MAX_PATH, szFilePath, &lpNameOnly);
		lstrcpy(lpNameOnly, TEXT("*.inf"));
	}
	else /* use uninstall file path */
	{
		GetFullPathName(lpUninstFile, MAX_PATH, szFilePath, &lpNameOnly);
	}
	
	hSearch = FindFirstFile(szFilePath, &searchData);
	if (hSearch == INVALID_HANDLE_VALUE)
	{ /* no .inf file(s) were found */
		if (!bQuiet)
		{
			LOADRESSTRING(3 + bUninstall);
			MessageBox(NULL, szResource, lpAppTitle, MB_OK | MB_ICONEXCLAMATION);
		}
		bErrors = TRUE;
		return;
	}
	
	do /* repeat for each file */
	{
	/* try to get the application's name from the .inf file ("SetupName") */
		lstrcpy(lpNameOnly, searchData.cFileName);
		GetPrivateProfileString(TEXT("Strings"), TEXT("SetupName"), searchData.cFileName, szProfileString, MAX_STRING, szFilePath);
		if (!szProfileString[0])
		{
			lstrcpy(szProfileString, searchData.cFileName);
		}
		
		if (!bQuiet)
		{ /* ask the user for permission to (un)install */
			LOADRESSTRING(5 + bUninstall);
			lpQuestion = (LPTSTR)malloc(lstrlen(szResource) + lstrlen(szProfileString));
			wsprintf(lpQuestion, szResource, szProfileString);
			iResult = MessageBox(NULL, lpQuestion, lpAppTitle, ((bUninstall)?MB_YESNO:MB_YESNOCANCEL) | MB_ICONQUESTION);
			free(lpQuestion);
		}
		else /* (un)install quietly */
		{
			iResult = IDYES;
		}
		
		if ((iResult == IDYES) && !bUninstall)
		{ /* install this file */
			lpQuestion = (LPTSTR)malloc(MAX_PATH);
		/* get "Program Files" path */
			RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("SOFTWARE\\Microsoft\\Windows\\CurrentVersion"), 0, KEY_READ, &hKey);
			RegQueryValueEx(hKey, TEXT("ProgramFilesDir"), NULL, NULL, (LPBYTE)lpQuestion, &dwSize);
			RegCloseKey(hKey);
		/* get the suggested subfolder (if available) */
			GetPrivateProfileString(TEXT("Strings"), TEXT("SetupBrowse"), TEXT(""), szProfileString, MAX_STRING, szFilePath);
			
			if (szProfileString[0])
			{ /* if the .inf file offers "SetupBrowse", it'll get "SetupDir" in return (set by the user) */
				if (bQuiet || BrowseForFolder(lpQuestion, szProfileString))
				{
					if (!WritePrivateProfileString(TEXT("Strings"), TEXT("SetupDir"), lpQuestion, szFilePath))
					{ /* display error message */
						if (!bQuiet)
						{
							LOADRESSTRING(12);
							lstrcpy(lpQuestion, szResource);
							wsprintf(szResource, lpQuestion, searchData.cFileName);
							MessageBox(NULL, szResource, lpAppTitle, MB_OK | MB_ICONEXCLAMATION);
						}
						bErrors = TRUE;
						iResult = IDNO;
					}
				}
				else /* abort installation */
				{
					iResult = IDNO;
				}
			}
			if (iResult == IDYES)
			{
				InstallINF(szFilePath, FALSE);
				if (szProfileString[0])
				{ /* remove the "SetupDir" entry, so that the .inf file remains portable */
					WritePrivateProfileString(TEXT("Strings"), TEXT("SetupDir"), TEXT(""), szFilePath);
				}
			}
			free(lpQuestion);
		}
		else if (iResult == IDYES)
		{ /* uninstall this file */
			InstallINF(szFilePath, TRUE);
			if (!bErrors)
			{ /* delete also Setup.exe */
				SelfDestruct();
			}
			iResult = IDCANCEL; /* uninstall only one file */
		}
	} while ((iResult != IDCANCEL) && FindNextFile(hSearch, &searchData));
	FindClose(hSearch);
}

UINT_PTR CALLBACK BrowseDialogProc(HWND hDlg, UINT uiMsg, WPARAM wParam, LPARAM lParam)
{
	TCHAR szPath[MAX_PATH];
	LONG i, lSelStart;
	BOOL bPathCorrected = FALSE;
	
	if (uiMsg == wm_LBSelChString)
	{ /* adjust the text label at the bottom of the dialog when the path changes */
		GetDlgItemText(hDlg, stc1, szPath, MAX_PATH);
		if (szResource[0] != 0)
		{ /* append the new subfolder's name */
			APPENDSLASH(szPath);
			lstrcat(szPath, szResource);
		}
		else if (szPath[3] == 0)
		{ /* always add a slash directly after the drive letter */
			APPENDSLASH(szPath);
		}
		SetDlgItemText(hDlg, stc1, szPath);
	}
	
	switch (uiMsg)
	{
	case WM_INITDIALOG:
		SendDlgItemMessage(hDlg, IDT_EDITSUBFOLDER, EM_SETLIMITTEXT, (WPARAM)128, 0);
		SetDlgItemText(hDlg, IDT_EDITSUBFOLDER, szResource);
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDT_EDITSUBFOLDER:
			if (HIWORD(wParam) == EN_UPDATE)
			{
				if (szResource[0] != 0)
				{ /* remove the subfolder's path */
					GetDlgItemText(hDlg, stc1, szPath, MAX_PATH);
					szPath[lstrlen(szPath) - lstrlen(szResource) - 1] = 0;
					SetDlgItemText(hDlg, stc1, szPath);
				}
				
			/* check the subfolder's name for unwanted characters */
				GetDlgItemText(hDlg, IDT_EDITSUBFOLDER, szResource, 129);
				SendDlgItemMessage(hDlg, IDT_EDITSUBFOLDER, EM_GETSEL, (WPARAM)&lSelStart, 0);
				for (i = 0; szResource[i] != 0; i++)
				{
					if (!IsCharAlphaNumeric(szResource[i]))
					{
						switch (szResource[i])
						{
						case '_': case '-': case '+': case '&': case '(': case ')': case '!': case '\'': case '~':
							break;
						case '.': case '\\':
							if (szResource[i] != szResource[i + 1])
							{ /* ".." and "\\" aren't allowed */
								break;
							}
						default: /* don't allow the rest */
							lstrcpy(&szResource[i], &szResource[i + 1]);
							if (lSelStart > i)
							{
								lSelStart--;
							}
							bPathCorrected = TRUE;
							break;
						}
					}
				}
				
			/* append the modified subfolder name */
				BrowseDialogProc(hDlg, wm_LBSelChString, 0, 0);
				if (bPathCorrected)
				{
					SetDlgItemText(hDlg, IDT_EDITSUBFOLDER, szResource);
					SendDlgItemMessage(hDlg, IDT_EDITSUBFOLDER, EM_SETSEL, (WPARAM)lSelStart, (LPARAM)lSelStart);
				}
			}
			break;
		}
		break;
	}
	
	return FALSE;
}

BOOL BrowseForFolder(LPTSTR lpInitFolder, LPTSTR lpAppSubFolder)
{
	BOOL bSuccess;
	OPENFILENAME ofn;
	LPTSTR lpFilter = TEXT("-\0!@~$@\0\0");
	TCHAR szDlgTitle[MAX_STRING], szFileName[MAX_PATH];
	
	LOADRESSTRING(8);
	lstrcpy(szDlgTitle, szResource);
	lstrcpy(szResource, lpAppSubFolder);
	szFileName[0] = 0;
	
BrowseRetryPoint:
	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hInstance = hInst;
	ofn.lpstrInitialDir = lpInitFolder;
	ofn.lpstrFile = szFileName;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrFilter = lpFilter;
	ofn.lpstrTitle = szDlgTitle;
	ofn.lpTemplateName = MAKEINTRESOURCE(IDD_BROWSEDLG);
	ofn.lpfnHook = BrowseDialogProc;
	ofn.lCustData = (LPARAM)lpAppSubFolder;
	ofn.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_ENABLETEMPLATE | OFN_LONGNAMES | OFN_ENABLEHOOK;
	
	if (bSuccess = GetOpenFileName(&ofn))
	{ /* cut off the (bogus) filename */
		szFileName[ofn.nFileOffset] = 0;
		GetShortPathName(szFileName, lpInitFolder, MAX_PATH);
		if (szResource[0] != 0)
		{ /* append the new subfolder's name */
			APPENDSLASH(lpInitFolder);
			lstrcat(lpInitFolder, szResource);
		}
	}
	else /* beep if there was an error */
	{
		switch (CommDlgExtendedError())
		{
		case 0:
			break;
		case FNERR_INVALIDFILENAME:
			lstrcpy(lpInitFolder, TEXT(""));
			goto BrowseRetryPoint;
		default:
			BEEP;
			bErrors = TRUE;
			break;
		}
	}
	
	return bSuccess;
}

VOID InstallINF(LPTSTR lpFileName, BOOL bUninst)
{
	TCHAR szInstallCommand[MAX_PATH + 48];
    SHELLEXECUTEINFO sei;
	
	wsprintf(szInstallCommand, TEXT("setupapi,InstallHinfSection %s 132 %s"), (bUninst)?TEXT("DefaultUninstall"):TEXT("DefaultInstall"), lpFileName);
	
	ZeroMemory(&sei, sizeof(sei));
	sei.cbSize = sizeof(sei);
	sei.fMask = (bUninst)?0:SEE_MASK_NOCLOSEPROCESS;
	sei.lpFile = TEXT("rundll32.exe");
	sei.lpParameters = szInstallCommand;
	sei.nShow = SW_NORMAL;
	
	if (ShellExecuteEx(&sei))
	{ /* wait until the (un)installation terminates */
		WaitForSingleObject(sei.hProcess, INFINITE);
		CloseHandle(sei.hProcess);
	}
	else /* notify the user */
	{
		if (!bQuiet)
		{
			LOADRESSTRING(13);
			MessageBox(NULL, szResource, lpAppTitle, MB_OK | MB_ICONEXCLAMATION);
		}
		bErrors = TRUE;
	}
}

VOID SelfDestruct(VOID)
{
	TCHAR szModule[MAX_PATH], szBatch[MAX_PATH], szModulePath[MAX_PATH];
	LPTSTR lpFileName, lpBatchContent;
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	HANDLE hFile;
	DWORD dwSize;
	
/* Get the paths to Setup.exe, the batch file and Setup's folder */
	GetModuleFileName(hInst, szModule, MAX_PATH);
	GetFullPathName(szModule, MAX_PATH, szBatch, &lpFileName);
	lstrcpy(lpFileName, TEXT("..\\TmpDlSp!.bat"));
	GetFullPathName(szBatch, MAX_PATH, szBatch, &lpFileName);
	lstrcpy(szModulePath, szModule);
	GetFullPathName(szModule, MAX_PATH, szModulePath, &lpFileName);
	*(lpFileName - 1) = 0;
	
	hFile = CreateFile(szBatch, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE)
	{ /* write and execute the batch file */
		LOADRESSTRING(10);
		lpBatchContent = (LPTSTR)malloc(4 * MAX_PATH + lstrlen(szResource));
		wsprintf(lpBatchContent, szResource, szModule, szModule, szModulePath, szBatch);
		dwSize = lstrlen(lpBatchContent);
		WriteFile(hFile, lpBatchContent, dwSize, &dwSize, NULL);
		CloseHandle(hFile);
		free(lpBatchContent);
		
		lstrcpy(szResource, TEXT("\""));
		lstrcat(szResource, szBatch);
		lstrcat(szResource, TEXT("\""));
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);
		ZeroMemory(&pi, sizeof(pi));
		CreateProcess(NULL, szResource, NULL, NULL, FALSE, IDLE_PRIORITY_CLASS, NULL, NULL, &si, &pi);
	}
	else /* couldn't delete Setup.exe */
	{
		if (!bQuiet)
		{
			LOADRESSTRING(14);
			lstrcpy(szBatch, szResource);
			wsprintf(szResource, szBatch, szModulePath);
			MessageBox(NULL, szResource, lpAppTitle, MB_OK | MB_ICONEXCLAMATION);
		}
		bErrors = TRUE;
	}
}
